/*
    jvm simulator for jvmpi-profiler testing
    Copyright (C) 2003 Robert Olofsson
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
    Robert Olofsson
    robo@khelekore.org
*/
/*
    This simulator is heavily influenced by the simulator.cpp found in
    Mike's Java Profiler
    Copyright (C) 2002  Michael J. Bresnahan
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <jni.h>
#include <jvmpi.h>
#include <pthread.h>
#include <jvm.h>

static JVMPI_Interface* jvmpi;

struct _JVMPI_RawMonitor {
    char* name;
    pthread_mutex_t mut;
};

jint JNICALL GetEnv (JavaVM *vm, void **penv, jint version) {
    fprintf (stderr, "GetEnv (%p, %p, %d)\n", vm, penv, version);
    *penv = (void*)jvmpi;
    return 0;
}

JVMPI_RawMonitor RawMonitorCreate (char *lock_name) {
    JVMPI_RawMonitor r;
    fprintf (stderr, "RawMonitorCreate (%s)\n", lock_name);
    r = (JVMPI_RawMonitor)malloc (sizeof (*r));
    r->name = strdup (lock_name);
    pthread_mutex_init (&r->mut, NULL);
    return r;
}

void RawMonitorEnter (JVMPI_RawMonitor r) {
    fprintf (stderr, "RawMonitorEnter (%s)\n", r->name);
    pthread_mutex_lock (&r->mut);
}

void RawMonitorExit (JVMPI_RawMonitor r) {
    fprintf (stderr, "RawMonitorExit (%s)\n", r->name);
    pthread_mutex_unlock (&r->mut);
}

void RawMonitorDestroy (JVMPI_RawMonitor r) {
    fprintf (stderr, "RawMonitorDestroy (%s)\n", r->name);
    if (r) {
	pthread_mutex_destroy (&r->mut);
	free (r->name);
	free (r);
    }
}

jint EnableEvent (jint event_type, void *arg) {
    fprintf (stderr, "EnableEvent (%d, %p)\n", event_type, arg);
    return JVMPI_SUCCESS;
}
jint DisableEvent (jint event_type, void *arg) {
    fprintf (stderr, "DisableEvent (%d, %p)\n", event_type, arg);
    return JVMPI_SUCCESS;
}
jint RequestEvent (jint event_type, void *arg) {
    fprintf (stderr, "RequestEvent (%d, %p)\n", event_type, arg);
    return JVMPI_SUCCESS;
}

void DisableGC (void) {
    fprintf (stderr, "DisableGC ()\n");
}
void EnableGC (void) {
    fprintf (stderr, "EnableGC ()\n");
}
void RunGC (void) {
    fprintf (stderr, "RunGC ()\n");
    call_gc_start ((JNIEnv*)pthread_self ());
    call_gc_finish ((JNIEnv*)pthread_self (), 4711, 
		    4711 * 8, 16 * 1024 * 1024);
}

struct local_storage; 
struct local_storage {
    JNIEnv* env_id;
    void*   ptr;
    struct local_storage* next;
};

struct local_storage* store = NULL;

void SetThreadLocalStorage (JNIEnv *env_id, void *ptr) {
    struct local_storage* ls = store;
    fprintf (stderr, "SetThreadLocalStorage (%p, %p)\n", env_id, ptr);
    while (ls != NULL) {
	if (ls->env_id == env_id) {
	    ls->ptr = ptr;
	    return;
	}
	ls = ls->next;
    }
    ls = (struct local_storage*)malloc (sizeof (*ls));
    ls->env_id = env_id;
    ls->ptr = ptr;
    ls->next = store;
    store = ls;
}

void* GetThreadLocalStorage (JNIEnv *env_id) {
    struct local_storage* ls = store;
    fprintf (stderr, "GetThreadLocalStorage (%p)\n", env_id);
    while (ls != NULL) {
	if (ls->env_id == env_id) {
	    return ls->ptr;
	}
	ls = ls->next;
    }
    return NULL;
}

static void FreeThreadLocalStorage () {
    struct local_storage* ls = store;
    struct local_storage* next;
    while (ls) {
	next = ls->next;
	free (ls);
	ls = next;
    }
}

static jint CreateSystemThread (char* name, jint priority, void (*f)(void*)) {
    pthread_t thread;
    int status;
    status = pthread_create (&thread, NULL, (void*)(*f), NULL);
    if (status == 0)
	return JNI_OK;
    return JNI_ERR;
}

void call_method_entry (JNIEnv* thread_env_id, jmethodID mid) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_METHOD_ENTRY;
    event.env_id = thread_env_id;
    event.u.method.method_id = mid;
    jvmpi->NotifyEvent (&event);
}

void call_method_exit (JNIEnv* thread_env_id, jmethodID mid) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_METHOD_EXIT;
    event.env_id = thread_env_id;
    event.u.method.method_id = mid;
    jvmpi->NotifyEvent (&event);
}

void call_object_alloc (JNIEnv* thread_env_id, jint arena_id, 
			jobjectID class_id, jint is_array, 
			jint size, jobjectID obj_id) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_OBJECT_ALLOC;
    event.env_id = thread_env_id;
    event.u.obj_alloc.arena_id = arena_id;
    event.u.obj_alloc.class_id = class_id;
    event.u.obj_alloc.is_array = is_array;
    event.u.obj_alloc.size = size;
    event.u.obj_alloc.obj_id = obj_id;
    jvmpi->NotifyEvent (&event);    
}

void call_object_move (JNIEnv* thread_env_id, 
		       jint arena_id, jobjectID obj_id, 
		       jint new_arena_id, jobjectID new_obj_id) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_OBJECT_MOVE;
    event.u.obj_move.arena_id = arena_id;
    event.u.obj_move.obj_id = obj_id;
    event.u.obj_move.new_arena_id = new_arena_id;
    event.u.obj_move.new_obj_id = new_obj_id;    
    jvmpi->NotifyEvent (&event);    
}
			       
void call_object_free (JNIEnv* thread_env_id, 
		       jobjectID obj_id) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_OBJECT_FREE;
    event.u.obj_free.obj_id = obj_id;
    jvmpi->NotifyEvent (&event);    
}

void call_gc_start (JNIEnv* thread_env_id) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_GC_START;
    event.env_id = thread_env_id;
    jvmpi->NotifyEvent (&event); 
}

void call_gc_finish (JNIEnv* thread_env_id, 
		     jlong used_objects, 
		     jlong used_object_space, 
		     jlong total_object_space) {
    JVMPI_Event event;
    event.event_type = JVMPI_EVENT_GC_FINISH;
    event.env_id = thread_env_id;
    event.u.gc_info.used_objects = used_objects;
    event.u.gc_info.used_object_space = used_object_space;
    event.u.gc_info.total_object_space = total_object_space;
    jvmpi->NotifyEvent (&event);     
}

int main (int argc, char **argv) {
    void* handle;
    char* error;
    char* options = NULL;
    
    struct JNIInvokeInterface_* jvm;
    JNIEXPORT jint JNICALL (*onload)(JavaVM *, char *, void *);

    if (argc < 2) {
	fprintf (stderr, "must give library to test...\n");
	exit (-1);
    }
    if (argc > 2) {
	options = argv[2];
    }

    handle = dlopen (argv[1], RTLD_LAZY);
    if (!handle) {
	fputs (dlerror (), stderr);
	exit(1);
    }

    onload = dlsym (handle, "JVM_OnLoad");
    if ((error = dlerror ()) != NULL)  {
	fprintf (stderr, "%s\n", error);
	exit(1);
    }
    
    jvm = (struct JNIInvokeInterface_*)malloc (sizeof (*jvm));
    jvm->GetEnv = GetEnv;
    jvmpi = (JVMPI_Interface*)malloc (sizeof (*jvmpi));
    jvmpi->RawMonitorCreate = RawMonitorCreate;
    jvmpi->RawMonitorEnter = RawMonitorEnter;
    jvmpi->RawMonitorExit = RawMonitorExit;
    jvmpi->RawMonitorDestroy = RawMonitorDestroy;
    jvmpi->EnableEvent = EnableEvent;
    jvmpi->DisableEvent = DisableEvent;
    jvmpi->RequestEvent = RequestEvent;
    jvmpi->DisableGC = DisableGC;
    jvmpi->EnableGC = EnableGC;
    jvmpi->RunGC = RunGC;
    jvmpi->SetThreadLocalStorage = SetThreadLocalStorage;
    jvmpi->GetThreadLocalStorage = GetThreadLocalStorage;
    jvmpi->CreateSystemThread = CreateSystemThread;
    
    fprintf (stderr, "calling JVM_OnLoad (...)\n");
    onload ((JavaVM*)&jvm, options, NULL);
    fprintf (stderr, "returned from JVM_OnLoad (...)\n");

    /* this needs lots of jni for jmp testing...*/
    {
	JVMPI_Event event;
	event.event_type = JVMPI_EVENT_JVM_INIT_DONE;
	jvmpi->NotifyEvent (&event);	
    }
    
    /** start a thread and end it. */
    {
	/** start a thread.. */
	JVMPI_Event event;
	JNIEnv* thread_env_id;
	char* class_name = "foo";
	char* source_name = "foo.java";
	int i = 0;
	int j = 0;
	fprintf (stderr, "starting a thread...\n");
	thread_env_id = (JNIEnv*)malloc (sizeof (*thread_env_id));

	event.event_type = JVMPI_EVENT_THREAD_START;
	event.env_id = NULL;
	event.u.thread_start.thread_name = "thread1";
	event.u.thread_start.group_name = "group1";
	event.u.thread_start.parent_name = "parent1";
	event.u.thread_start.thread_id = NULL;
	event.u.thread_start.thread_env_id = thread_env_id;
	jvmpi->NotifyEvent (&event);

	
	/** Test class load, (method_entry, method_exit)*X, class unload... */
	fprintf (stderr, "loading a class...\n");
	event.event_type = JVMPI_EVENT_CLASS_LOAD;
	event.env_id = thread_env_id;
	event.u.class_load.class_name = class_name;
	event.u.class_load.source_name = source_name;
	event.u.class_load.num_interfaces = 0;
	event.u.class_load.num_methods = 4;
	event.u.class_load.methods = (JVMPI_Method*)malloc (sizeof (JVMPI_Method) * event.u.class_load.num_methods);
	event.u.class_load.methods[i].method_id = (jmethodID)i;
	event.u.class_load.methods[i].method_name = "a";
	event.u.class_load.methods[i].start_lineno = 0;
	event.u.class_load.methods[i].end_lineno = 0;
	event.u.class_load.methods[i++].method_signature = "()V";	
	event.u.class_load.methods[i].method_id = (jmethodID)i;
	event.u.class_load.methods[i].method_name = "b";
	event.u.class_load.methods[i].start_lineno = 0;
	event.u.class_load.methods[i].end_lineno = 0;
	event.u.class_load.methods[i++].method_signature = "()V";
	event.u.class_load.methods[i].method_id = (jmethodID)i;
	event.u.class_load.methods[i].method_name = "c";
	event.u.class_load.methods[i].start_lineno = 0;
	event.u.class_load.methods[i].end_lineno = 0;
	event.u.class_load.methods[i++].method_signature = "(IJZF)V";
	event.u.class_load.methods[i].method_id = (jmethodID)i;
	event.u.class_load.methods[i].method_name = "d";
	event.u.class_load.methods[i].start_lineno = 0;
	event.u.class_load.methods[i].end_lineno = 0;
	event.u.class_load.methods[i++].method_signature = "(II[I)V";
	event.u.class_load.num_static_fields = 0;
	event.u.class_load.statics = NULL;
	event.u.class_load.num_instance_fields = 0;
	event.u.class_load.instances = NULL;
	event.u.class_load.class_id = 0;
	jvmpi->NotifyEvent (&event); 
	
	free (event.u.class_load.methods);
	
	fprintf (stderr, "doing some method_entry, method_exit...\n");
	event.env_id = thread_env_id;
	for (i = 0; i < 10; i++) {	   
	    call_method_entry (thread_env_id, (jmethodID)0);
	    for (j = 0; j < 100; j++) {
		call_object_alloc (thread_env_id, 0, (jobjectID)0, 0, 8, (jobjectID)((i << 16) + j));
	    }
	    
	    for (j = 0; j < 10; j++) {
		call_method_entry (thread_env_id, (jmethodID)1);
		if (j < 6) {
		    call_method_entry (thread_env_id, (jmethodID)2);
		    call_method_exit (thread_env_id, (jmethodID)2);
 		}
		call_method_exit (thread_env_id, (jmethodID)1);
		if (j > 4) {
		    call_method_entry (thread_env_id, (jmethodID)3);
		    call_method_exit (thread_env_id, (jmethodID)3);
		}
	    }	
	    call_method_exit (thread_env_id, (jmethodID)0);
	}

	call_gc_start (thread_env_id);
	call_object_move (thread_env_id, 0, (jobjectID)0, 0, (jobjectID)1000);
	for (j = 1; j < 100; j++) {
	    call_object_free (thread_env_id, (jobjectID)j);
	}
	call_gc_finish (thread_env_id, 901, 7208, 16 * 1024 * 1024);
	
	fprintf (stderr, "unloading class...\n");
	event.event_type = JVMPI_EVENT_CLASS_UNLOAD;
	event.env_id = thread_env_id;
	event.u.class_unload.class_id = 0;		
	jvmpi->NotifyEvent (&event);

	/** end the thread.. */
	fprintf (stderr, "ending thread...\n");
	event.env_id = thread_env_id;
	event.event_type = JVMPI_EVENT_THREAD_END;
	jvmpi->NotifyEvent (&event);
	
	free (thread_env_id);
	fprintf (stderr, "simple thread test ok...\n");
    }

    sleep (1000);
    /** Send the shut down. */
    {
	JVMPI_Event event;
	fprintf (stderr, "sending shut down event to close down...\n");
	event.event_type = JVMPI_EVENT_JVM_SHUT_DOWN;
	event.env_id = NULL;
	jvmpi->NotifyEvent (&event);	
	fprintf (stderr, "shut down sent...\n");
    }

    fprintf (stderr, "freeing simulator structures...\n");
    FreeThreadLocalStorage ();
    free (jvmpi);
    free (jvm);
    dlclose(handle);
    fprintf (stderr, "exiting...\n");
    return 0;
}
